@@ -22,19 +22,30 @@ server. All commands listed here should have counterparts in the
 launch_control.dashboard_app.xml_rpc package.
 """
 
+import argparse
+import contextlib
 import errno
 import os
+import re
 import socket
 import sys
-import urlparse
 import urllib
+import urlparse
 import xmlrpclib
 
-import argparse
-
 from launch_control_tool.interface import Command
 
 
+class InsufficientServerVersion(Exception):
+    """
+    Exception raised when server version that a command interacts with is too
+    old to support required features.
+    """
+    def __init__(self, server_version, required_version):
+        self.server_version = server_version
+        self.required_version = required_version
+
+
 class DataSetRenderer(object):
     """
     Support class for rendering a table out of list of dictionaries.
@@ -279,6 +290,21 @@ class XMLRPCCommand(Command):
         return urlparse.urlunsplit(
             (parts.scheme, parts.netloc, parts.path.rstrip("/") + "/xml-rpc/", "", ""))
 
+    def _check_server_version(self, server_obj, required_version):
+        """
+        Check that server object has is at least required_version.
+        which must be a tuple of integer (major, minor, micro).
+
+        This method may raise InsufficientServerVersion.
+        """
+        from distutils.version import StrictVersion, LooseVersion
+        # For backwards compatibility the server reports
+        # major.minor.micro.releaselevel.serial which is not PEP-386 compliant
+        server_version = LooseVersion(server_obj.version())
+        required_version = LooseVersion(required_version)
+        if server_version < required_version:
+            raise InsufficientServerVersion(server_version, required_version)
+
     def __init__(self, parser, args):
         super(XMLRPCCommand, self).__init__(parser, args)
         xml_rpc_url = self._construct_xml_rpc_url(self.args.dashboard_url) 
@@ -291,7 +317,7 @@ class XMLRPCCommand(Command):
         default_dashboard_url = os.getenv("DASHBOARD_URL")
         if default_dashboard_url:
             group.add_argument("--dashboard-url",
-                    metavar="URL", help="URL of your validation dashboard",
+                    metavar="URL", help="URL of your validation dashboard (%(default)s)",
                     default=default_dashboard_url)
         else:
             group.add_argument("--dashboard-url", required=True,
@@ -301,9 +327,10 @@ class XMLRPCCommand(Command):
                 help="Show XML-RPC data")
         return group
 
-    def invoke(self):
+    @contextlib.contextmanager
+    def safety_net(self):
         try:
-            return self.invoke_remote()
+            yield
         except socket.error as ex:
             print >>sys.stderr, "Unable to connect to server at %s" % (
                     self.args.dashboard_url,)
@@ -324,7 +351,14 @@ class XMLRPCCommand(Command):
             print >>sys.stderr, "HTTP error code: %d/%s" % (ex.errcode, ex.errmsg)
         except xmlrpclib.Fault as ex:
             self.handle_xmlrpc_fault(ex.faultCode, ex.faultString)
-        return -1
+        except InsufficientServerVersion as ex:
+            print >>sys.stderr, ("This command requires at least server version "
+                                 "%s, actual server version is %s" %
+                                 (ex.required_version, ex.server_version))
+
+    def invoke(self):
+        with self.safety_net():
+            return self.invoke_remote()
 
     def handle_xmlrpc_fault(self, faultCode, faultString):
         if faultCode == 500:
@@ -337,6 +371,61 @@ class XMLRPCCommand(Command):
         raise NotImplementedError()
 
 
+class ExperimentalNoticeAction(argparse.Action):
+    """
+    Argparse action that implements the --experimental-notice
+    """
+
+    message = """
+    Some lc-tool sub-commands are marked as EXPERIMENTAL. Those commands are
+    not guaranteed to work identically, or have identical interface between
+    subsequent lc-tool releases.
+
+    We do that to make it possible to provide good user interface and
+    server-side API when working on new features. Once a feature is stabilized
+    the UI will be frozen and all subsequent changes will retain backwards
+    compatibility.
+    """
+    message = message.lstrip()
+    message = re.sub(re.compile("[ \t]+", re.M), " ", message)
+    message = re.sub(re.compile("^ ", re.M), "", message)
+
+    def __init__(self,
+                 option_strings, dest, default=None, required=False,
+                 help=None):
+        super(ExperimentalNoticeAction, self).__init__(
+            option_strings=option_strings, dest=dest, default=default, nargs=0,
+            help=help)
+
+    def __call__(self, parser, namespace, values, option_string=None):
+        parser.exit(message=self.message)
+
+
+class ExperimentalCommandMixIn(object):
+    """
+    Experimental command.
+
+    Prints a warning message on each call to invoke()
+    """
+
+    def invoke(self):
+        self.print_experimental_notice()
+        return super(ExperimentalCommandMixIn, self).invoke()
+
+    @classmethod
+    def register_arguments(cls, parser):
+        retval = super(ExperimentalCommandMixIn, cls).register_arguments(parser)
+        parser.register("action", "experimental_notice", ExperimentalNoticeAction)
+        parser.add_argument("--experimental-notice",
+                            action="experimental_notice",
+                            default=argparse.SUPPRESS,
+                            help="Explain the nature of experimental commands")
+        return retval
+    
+    def print_experimental_notice(self):
+        print "EXPERIMENTAL - SUBJECT TO CHANGE (See --experimental-notice for more info)"
+
+
 class server_version(XMLRPCCommand):
     """
     Display dashboard server version
@@ -524,18 +613,30 @@ class make_stream(XMLRPCCommand):
     @classmethod
     def register_arguments(cls, parser):
         super(make_stream, cls).register_arguments(parser)
-        parser.add_argument("pathname",
-                type=str,
-                help="Pathname of the bundle stream to create")
+        parser.add_argument(
+            "pathname",
+            type=str,
+            help="Pathname of the bundle stream to create")
+        parser.add_argument(
+            "--name",
+            type=str,
+            default=None,
+            help="Name of the bundle stream (description)")
+
 
     def invoke_remote(self):
-        pathname = self.server.make_stream(self.args.pathname)
+        self._check_server_version(self.server, "0.3.0.candidate.9")
+        pathname = self.server.make_stream(self.args.pathname, self.args.name)
         print "Bundle stream {pathname} created".format(pathname=pathname)
 
 
 class backup(XMLRPCCommand):
     """
-    Backup a dashboard instance
+    Backup data uploaded to a dashboard instance.
+    
+    Not all data is preserved. The following data is lost: identity of the user
+    that uploaded each bundle, time of uploading and deserialization on the
+    server, name of the bundle stream that contained the data
     """
 
     @classmethod
@@ -572,13 +673,14 @@ class restore(XMLRPCCommand):
                             help="Directory to backup from")
 
     def invoke_remote(self):
+        self._check_server_version(self.server, "0.3.0.candidate.9")
         for stream_pathname_quoted in os.listdir(self.args.BACKUP_DIR):
             filesystem_stream_pathname = os.path.join(self.args.BACKUP_DIR, stream_pathname_quoted)
             if not os.path.isdir(filesystem_stream_pathname):
                 continue
             stream_pathname = urllib.unquote(stream_pathname_quoted)
             print "Processing stream %s" % stream_pathname
-            self.server.make_stream(stream_pathname)
+            self.server.make_stream(stream_pathname, "Restored from backup")
             for content_filename_quoted in os.listdir(filesystem_stream_pathname):
                 filesystem_content_filename = os.path.join(filesystem_stream_pathname, content_filename_quoted)
                 if not os.path.isfile(filesystem_content_filename):
@@ -590,7 +692,7 @@ class restore(XMLRPCCommand):
                 self.server.put(content, content_filename, stream_pathname)
             
 
-class pull(XMLRPCCommand):
+class pull(ExperimentalCommandMixIn, XMLRPCCommand):
     """
     Pull bundles and bundle streams from one REMOTE DASHBOARD to DASHBOARD
     """
@@ -605,21 +707,159 @@ class pull(XMLRPCCommand):
     def register_arguments(cls, parser):
         group = super(pull, cls).register_arguments(parser)
         group.add_argument("--remote-dashboard-url", required=True,
-                metavar="REMOTE_URL", help="URL of the remote validation dashboard")
+                metavar="URL", help="URL of the remote validation dashboard")
+        group.add_argument("STREAM", nargs="*", help="Streams to pull from (all by default)")
+
+    @staticmethod
+    def _filesizeformat(num_bytes):
+        """
+        Formats the value like a 'human-readable' file size (i.e. 13 KB, 4.1 MB,
+        102 num_bytes, etc).
+        """
+        try:
+            num_bytes = float(num_bytes)
+        except (TypeError,ValueError,UnicodeDecodeError):
+            return "%(size)d byte", "%(size)d num_bytes" % {'size': 0}
+
+        filesize_number_format = lambda value: "%0.2f" % (round(value, 1),)
+
+        if num_bytes < 1024:
+            return "%(size)d bytes" % {'size': num_bytes}
+        if num_bytes < 1024 * 1024:
+            return "%s KB" % filesize_number_format(num_bytes / 1024)
+        if num_bytes < 1024 * 1024 * 1024:
+            return "%s MB" % filesize_number_format(num_bytes / (1024 * 1024))
+        return "%s GB" % filesize_number_format(num_bytes / (1024 * 1024 * 1024))
 
     def invoke_remote(self):
+        self._check_server_version(self.server, "0.3.0.candidate.9")
         print "Checking local and remote streams"
-        remote = [stream["pathname"] for stream in self.remote_server.streams()]
-        local = [stream["pathname"] for stream in self.server.streams()]
-        missing = set(remote) - set(local)
-        for stream_pathname in remote:
-            if stream_pathname in missing:
-                print "Creating missing stream %s" % stream_pathname
-                self.server.make_stream(stream_pathname)
-            local_bundles = [bundle["content_sha1"] for bundle in self.server.bundles(stream_pathname)]
-            remote_bundles = [bundle["content_sha1"] for bundle in self.remote_server.bundles(stream_pathname)]
-            missing_bundles = set(remote_bundles) - set(local_bundles)
+        remote = self.remote_server.streams()
+        if self.args.STREAM:
+            # Check that all requested streams are available remotely
+            requested_set = frozenset(self.args.STREAM)
+            remote_set = frozenset((stream["pathname"] for stream in remote))
+            unavailable_set = requested_set - remote_set
+            if unavailable_set:
+                print >>sys.stderr, "Remote stream not found: %s" % ", ".join(unavailable_set)
+                return -1
+            # Limit to requested streams if necessary
+            remote = [stream for stream in remote if stream["pathname"] in requested_set]
+        local = self.server.streams()
+        missing_pathnames = set([stream["pathname"] for stream in remote]) - set([stream["pathname"] for stream in local])
+        for stream in remote:
+            local_bundles = [bundle for bundle in self.server.bundles(stream["pathname"])]
+            remote_bundles = [bundle for bundle in self.remote_server.bundles(stream["pathname"])]
+            missing_bundles = set((bundle["content_sha1"] for bundle in remote_bundles)) - set((bundle["content_sha1"] for bundle in local_bundles))
+            try:
+                missing_bytes = sum((bundle["content_size"] for bundle in remote_bundles if bundle["content_sha1"] in missing_bundles))
+            except KeyError as ex:
+                # Older servers did not return content_size so this part is optional
+                missing_bytes = None
+            if missing_bytes:
+                print "Stream %s needs update (%s)" % (stream["pathname"], self._filesizeformat(missing_bytes))
+            elif missing_bundles:
+                print "Stream %s needs update" % (stream["pathname"],)
+            else:
+                print "Stream %s is up to date" % (stream["pathname"],)
+            if stream["pathname"] in missing_pathnames:
+                self.server.make_stream(stream["pathname"], stream["name"])
             for content_sha1 in missing_bundles:
-                print "Pulling bundle %s" % content_sha1
+                print "Getting %s" % (content_sha1,),
+                sys.stdout.flush()
                 data = self.remote_server.get(content_sha1)
-                self.server.put(data["content"], data["content_filename"], stream_pathname)
+                print "got %s, storing" % (self._filesizeformat(len(data["content"]))),
+                sys.stdout.flush()
+                self.server.put(data["content"], data["content_filename"], stream["pathname"])
+                print "done"
+
+
+class data_views(ExperimentalCommandMixIn, XMLRPCCommand):
+    """
+    Show data views defined on the server
+    """
+    renderer = DataSetRenderer(
+        column_map = {
+            'name': 'Name',
+            'summary': 'Summary',
+        },
+        order = ('name', 'summary'),
+        empty = "There are no data views defined yet",
+        caption = "Data Views",
+        separator = " | ")
+
+    def invoke_remote(self):
+        self._check_server_version(self.server, "0.4.0.dev")
+        self.renderer.render(self.server.data_views())
+        print
+        print "Tip: to invoke a data view try `lc-tool query-data-view`"
+
+
+class query_data_view(ExperimentalCommandMixIn, XMLRPCCommand):
+    """
+    Invoke a specified data view
+    """
+    @classmethod
+    def register_arguments(cls, parser):
+        super(query_data_view, cls).register_arguments(parser)
+        parser.add_argument("QUERY", metavar="QUERY", nargs="...",
+                           help="Data view name and any optional and required arguments")
+
+    def _probe_data_views(self):
+        """
+        Probe the server for information about data views
+        """
+        with self.safety_net():
+            self._check_server_version(self.server, "0.4.0.dev")
+            return self.server.data_views()
+
+    def reparse_arguments(self, parser, raw_args):
+        self.data_views = self._probe_data_views()
+        if self.data_views is None:
+            return
+        # Here we hack a little, the last actuin is the QUERY action added
+        # in register_arguments above. By removing it we make the output
+        # of lc-tool query-data-view NAME --help more consistent.
+        del parser._actions[-1]
+        subparsers = parser.add_subparsers(
+            title="Data views available on the server")
+        for data_view in self.data_views: 
+            data_view_parser = subparsers.add_parser(
+                data_view["name"],
+                help=data_view["summary"],
+                epilog=data_view["documentation"])
+            data_view_parser.set_defaults(data_view=data_view)
+            group = data_view_parser.add_argument_group("Data view parameters")
+            for argument in data_view["arguments"]:
+                group.add_argument(
+                    "--{name}".format(name=argument["name"]),
+                    help=argument["help"],
+                    type=str,
+                    default=argument["default"])
+        self.args = self.parser.parse_args(raw_args)
+
+    def invoke_remote(self):
+        if self.data_views is None:
+            return -1
+        self._check_server_version(self.server, "0.4.0.dev")
+        # Build a collection of arguments for data view
+        data_view_args = {}
+        for argument in self.args.data_view["arguments"]:
+            arg_name = argument["name"]
+            if arg_name in self.args:
+                data_view_args[arg_name] = getattr(self.args, arg_name) 
+        # Invoke the data view
+        response = self.server.query_data_view(self.args.data_view["name"], data_view_args) 
+        # Create a pretty-printer
+        renderer = DataSetRenderer(
+            separator=" | ",
+            caption=self.args.data_view["summary"],
+            order=[item["name"] for item in response["columns"]])
+        # Post-process the data so that it fits the printer
+        data_for_renderer = [
+            dict(zip(
+                [column["name"] for column in response["columns"]],
+                row))
+            for row in response["rows"]]
+        # Print the data
+        renderer.render(data_for_renderer)
